/*
    module/atmio-E.c
    Hardware driver for NI AT-MIO E series cards

    COMEDI - Linux Control and Measurement Device Interface
    Copyright (C) 1997-8 David A. Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*
	The real guts of the driver is in ni-E.c, which is included
	both here and in pcimio-E.c

	
	Interrupt support added by Truxton Fulton <trux@truxton.com>

	References for specifications:
	
	   321747b.pdf  Register Level Programmer Manual (obsolete)
	   321747c.pdf  Register Level Programmer Manual (new)
	   DAQ-STC reference manual

	Other possibly relevant info:
	
	   320517c.pdf  User manual (obsolete)
	   320517f.pdf  User manual (new)
	   320889a.pdf  delete
	   320906c.pdf  maximum signal ratings
	   321066a.pdf  about 16x
	   321791a.pdf  discontinuation of at-mio-16e-10 rev. c
	   321808a.pdf  about at-mio-16e-10 rev P
	   321837a.pdf  discontinuation of at-mio-16de-10 rev d
	   321838a.pdf  about at-mio-16de-10 rev N
	
	ISSUES:

	need to deal with external reference for DAC, and other DAC
	properties in board properties
	
	deal with at-mio-16de-10 revision D to N changes, etc.
	
*/

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#include <linux/malloc.h>
#include <comedi_module.h>
#include <ni_stc.h>

#undef DEBUG

#define ATMIO 1
#undef PCIMIO

/*
 *  AT specific setup
 */

#define NI_SIZE 0x20


static ni_board ni_boards[]={
	{44,"at-mio-16e-1",
		16,12,0xfff,0x800,
		8192,0,0,
		2,12,0xfff,0x800,
		2048,0,
		1,1 },
	{25,"at-mio-16e-2",
		16,12,0xfff,0x800,
		2048,0,0,
		2,12,0xfff,0x800,
		2048,0,
		1,1 },
	{36,"at-mio-16e-10",
		16,12,0xfff,0x800,
		512,0,0,
		2,12,0xfff,0x800,
		0,0,
		1,2 },
	{37,"at-mio-16de-10",
		16,12,0xfff,0x800,
		512,0,0,
		2,12,0xfff,0x800,
		0,0,
		1,2 },
	{38,"at-mio-64e-3",
		64,12,0xfff,0x800,
		2048,0,0,
		2,12,0xfff,0x800,
		2048,0,
		1,1 },
	{39,"at-mio-16xe-50",
		16,16,0xffff,0x8000,
		512,1,1,
		2,12,0xfff,0x800,
		0,1,
		2,3 },
	{00,"at-mio-16xe-10",	/* unknown id */
		16,16,0xffff,0x8000,
		512,1,0,
		2,12,0xfff,0x800,
		-1,0, /* don't know fifo depth */
		2,3 },
	{00,"at-ai-16xe-10",	/* unknown id */
		16,16,0xffff,0x8000,
		512,1,0,
		0,0,0,0,
		0,0,
		2,3 }
};


static int ni_irqpin[]={-1,-1,-1,0,1,2,-1,3,-1,-1,4,5,6,-1,-1,7};

#define interrupt_pin(a)	(ni_irqpin[(a)])



/* How we access registers */

#define ni_writew(a,b)		(outw((a),(b)+dev->iobase))
#define ni_readw(a)		(inw((a)+dev->iobase))
#define ni_writeb(a,b)		(outb((a),(b)+dev->iobase))
#define ni_readb(a)		(inb((a)+dev->iobase))
#define ni_writeb_p(a,b)	ni_writeb(a,b)
#define ni_readb_p(a)		ni_readb(a)


/*
 * this is how we access windowed registers
 */

#define win_out(a,b) (ni_writew((b),Window_Address),ni_writew((a),Window_Data))
#define win_in(b) (ni_writew((b),Window_Address),ni_readw(Window_Data))


typedef struct{
	int board;
	int dio;
	int ao0p,ao1p;
	int lastchan;
	int last_do;
}ni_private;
#define devpriv ((ni_private *)dev->private)

static atmio_E_rem(comedi_device *dev);
#define ni_E_rem_stage2 atmio_E_rem


#include "ni-E.c"


static int init_stage2(comedi_device *dev,comedi_devconfig *it);
static int ni_getboardtype(comedi_device *dev);

/* clean up allocated resources */
int atmio_E_free(comedi_device *dev)
{
	if(dev->iobase)
		release_region(dev->iobase,NI_SIZE);
	if(dev->irq)
		free_irq(dev->irq,dev);
	if(dev->private)
		kfree(dev->private);

	return 0;
}

/* called when driver is removed */
static int atmio_E_rem(comedi_device *dev)
{
	return atmio_E_free(dev);
}

int atmio_E_init(comedi_device *dev,comedi_devconfig *it)
{
	int ret;
	
	if(!strcmp("ni_E",it->board_name)){
		printk("comedi: 'ni_E' deprecated.  Use 'atmio-E'\n");
	}else if(!strcmp("atmio-E",it->board_name)){
		;
	}else{
		return 0;
	}

	if( !(ret=init_stage2(dev,it)) )
		return 1;

	atmio_E_free(dev);
	
	return ret;
}

static int init_stage2(comedi_device *dev,comedi_devconfig *it)
{
	int		ret;
	int		iobase;
	int		board;
	int		irq;

	
	/* reserve our I/O region */

	iobase=0x200;
	if(it->options[0])iobase=it->options[0];
	
	printk("comedi%d: ni_E: 0x%04x",dev->minor,iobase);
	if(check_region(iobase,NI_SIZE)<0){
		printk(" I/O port conflict\n");
		return -EIO;
	}
	request_region(iobase,NI_SIZE,"ni_E");

	dev->iobase=iobase;
	dev->iosize=NI_SIZE;
	
	
	/* board existence sanity check */
	
	
	/* get board type */

	board=ni_getboardtype(dev);
	if(board<0)return -EIO;
	
	printk(" %s",ni_boards[board].name);
	dev->board_name=ni_boards[board].name;
	dev->driver_name="atmio-E";


	/* irq stuff */

	irq=it->options[1];
	if(irq!=0){
		if(irq<0 || irq>15 || ni_irqpin[irq]==-1){
			printk(" invalid irq\n");
			return -EINVAL;
		}
		printk(" ( irq = %d )",irq);
		if( (ret=request_irq(irq,ni_E_interrupt,SA_INTERRUPT,"atmio-E",dev))<0 ){
			printk(" irq not available\n");
			return -EINVAL;
		}
		dev->irq=irq;
	}
	
	/* allocate private area */
	
	dev->private=kmalloc(sizeof(ni_private),GFP_KERNEL);
	if(!dev->private){
		return -ENOMEM;
	}
	memset(dev->private,0,sizeof(ni_private));
	
	/* generic E series stuff in ni-E.c */

	if( (ret=ni_E_init(dev,it,board))<0 ){
		return ret;
	}
	
	return 0;
}


static int ni_getboardtype(comedi_device *dev)
{
	int device_id=ni_read_eeprom(dev,511);
	int i;
	
	for(i=0;i<n_ni_boards;i++){
		if(ni_boards[i].device_id==device_id){
			return i;
		}
	}
	if(device_id==255){
		printk(" can't find board\n");
	}else{
		printk(" unknown device ID %d -- contact author\n",device_id);
	}
	return -1;
}


