/*
    module/module.c
    comedi kernel module

    COMEDI - Linux Control and Measurement Device Interface
    Copyright (C) 1997-8 David A. Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/



#include <comedi_module.h>

#ifdef LINUX_V22
#define EXPORT_SYMTAB
#endif

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fcntl.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <asm/io.h>
#include <asm/uaccess.h>

comedi_device *comedi_devices;


static int do_devconfig_ioctl(comedi_device *dev,comedi_devconfig *arg,kdev_t minor);
static int do_devinfo_ioctl(comedi_device *dev,comedi_devinfo *arg);
static int do_subdinfo_ioctl(comedi_device *dev,comedi_subdinfo *arg,void *file);
static int do_chaninfo_ioctl(comedi_device *dev,comedi_subdinfo *arg);
static int do_trig_ioctl(comedi_device *dev,void *arg,void *file);
static int do_lock_ioctl(comedi_device *dev,unsigned int arg,void * file);
static int do_unlock_ioctl(comedi_device *dev,unsigned int arg,void * file);

static void postconfig(comedi_device *dev);

static int comedi_ioctl(struct inode * inode,struct file * file,unsigned int cmd,unsigned long arg)
{
	kdev_t minor=MINOR(inode->i_rdev);
	comedi_device *dev=comedi_devices+minor;
	
	switch(cmd)
	{
	case COMEDI_DEVCONFIG:
		return do_devconfig_ioctl(dev,(void *)arg,minor);
	case COMEDI_DEVINFO:
		return do_devinfo_ioctl(dev,(void *)arg);
	case COMEDI_SUBDINFO:
		return do_subdinfo_ioctl(dev,(void *)arg,file);
	case COMEDI_CHANINFO:
		return do_chaninfo_ioctl(dev,(void *)arg);
	case COMEDI_TRIG:
		return do_trig_ioctl(dev,(void *)arg,file);
	case COMEDI_LOCK:
		return do_lock_ioctl(dev,arg,file);
	case COMEDI_UNLOCK:
		return do_unlock_ioctl(dev,arg,file);
	default:
		return -EIO;
	}
}


/*
	COMEDI_DEVCONFIG
	device config ioctl
	
	arg:
		pointer to devconfig structure
	
	reads:
		devconfig structure at arg
	
	writes:
		none
*/
static int do_devconfig_ioctl(comedi_device *dev,comedi_devconfig *arg,kdev_t minor)
{
	comedi_devconfig it;
	int ret,i;
	
	if(!suser())
		return -EPERM;
	
	if(copy_from_user(&it,arg,sizeof(comedi_devconfig)))
		return -EFAULT;
	
	if(strnlen(it.board_name,COMEDI_NAMELEN)==COMEDI_NAMELEN)
		return -EINVAL;
	
	if(dev->rem){
		ret=dev->rem(dev);
		if(ret<0)
			return ret;
	}
	memset(dev,0,sizeof(comedi_device));

	if(!it.board_name[0])return 0;
	
	dev->minor=minor;
	for(i=0;i<comedi_maxinits;i++){
		ret=comedi_devinits[i](dev,&it);
		if(ret<0)
			return ret;
		if(ret){
			/* do a little post-config cleanup */
			postconfig(dev);
			return 0;
		}
	}
	
	return -EINVAL;
}


/*
	COMEDI_DEVINFO
	device info ioctl
	
	arg:
		pointer to devinfo structure
	
	reads:
		none
	
	writes:
		devinfo structure
		
*/
static int do_devinfo_ioctl(comedi_device *dev,comedi_devinfo *arg)
{
	comedi_devinfo devinfo;
	
	
	/* fill devinfo structure */
	devinfo.version_code=COMEDI_VERSION_CODE;
	devinfo.n_subdevs=dev->n_subdevices;
	memcpy(devinfo.driver_name,dev->driver_name,COMEDI_NAMELEN);
	memcpy(devinfo.board_name,dev->board_name,COMEDI_NAMELEN);
	memcpy(devinfo.options,dev->options,COMEDI_NDEVCONFOPTS*sizeof(int));
	

	if(copy_to_user(arg,&devinfo,sizeof(comedi_devinfo)))
		return -EFAULT;

	return 0;
}


/*
	COMEDI_SUBDINFO
	subdevice info ioctl
	
	arg:
		pointer to array of subdevice info structures
	
	reads:
		none
	
	writes:
		array of subdevice info structures at arg
		
*/
static int do_subdinfo_ioctl(comedi_device *dev,comedi_subdinfo *arg,void *file)
{
	int ret,i;
	comedi_subdinfo *tmp,*us;
	comedi_subdevice *s;
	

	tmp=kmalloc(dev->n_subdevices*sizeof(comedi_subdinfo),GFP_KERNEL);
	if(!tmp)
		return -ENOMEM;
	
	/* fill subdinfo structs */
	for(i=0;i<dev->n_subdevices;i++){
		s=dev->subdevices+i;
		us=tmp+i;
		
		us->n_chan		= s->n_chan;
		us->subd_flags		= s->subdev_flags;
		us->timer_type		= s->timer_type;
		us->len_chanlist	= s->len_chanlist;
		us->maxdata		= s->maxdata;
		us->range_type		= s->range_type;
		
		if(s->busy)
			us->subd_flags |= SDF_BUSY;
		if(s->busy == file)
			us->subd_flags |= SDF_BUSY_OWNER;
		if(s->lock)
			us->subd_flags |= SDF_LOCKED;
		if(s->lock == file)
			us->subd_flags |= SDF_LOCK_OWNER;
	}
	
	ret=copy_to_user(arg,tmp,dev->n_subdevices*sizeof(comedi_subdinfo));
	
	kfree(tmp);
	
	return ret?-EFAULT:0;
}


/*
	COMEDI_CHANINFO
	subdevice info ioctl
	
	arg:
		pointer to chaninfo structure
	
	reads:
		chaninfo structure at arg
	
	writes:
		arrays at elements of chaninfo structure
	
*/
static int do_chaninfo_ioctl(comedi_device *dev,comedi_subdinfo *arg)
{
	comedi_subdevice *s;
	comedi_chaninfo it;
	int ret;
	
	if(copy_from_user(&it,arg,sizeof(comedi_chaninfo)))
		return -EFAULT;
	
	if(it.subdev>=dev->n_subdevices)
		return -EINVAL;
	s=dev->subdevices+it.subdev;
	
	if(it.flaglist){
		if(s->subdev_flags & SDF_FLAGS)
			ret=copy_to_user(it.flaglist,s->flaglist,s->n_chan*sizeof(unsigned int));
		else
			ret=clear_user(it.flaglist,s->n_chan*sizeof(unsigned int));
		if(ret)return -EFAULT;
	}
			
	if(it.rangelist){
		if(s->subdev_flags & SDF_FLAGS)
			ret=copy_to_user(it.rangelist,s->range_list,s->n_chan*sizeof(unsigned int));
		else
			ret=clear_user(it.rangelist,s->n_chan*sizeof(unsigned int));
		if(ret)return -EFAULT;
	}
	
	return 0;
}


/*
	COMEDI_TRIG
	trigger ioctl
	
	arg:
		pointer to trig structure
	
	reads:
		trig structure at arg
		channel/range list
	
	writes:
		modified trig structure at arg
		data list

*/
static int do_trig_ioctl(comedi_device *dev,void *arg,void *file)
{
	comedi_trig it;
	comedi_subdevice *s;
	unsigned int *user_chanlist;
	unsigned int *chanlist=NULL;
	int ret=0,i;
	
	if(copy_from_user(&it,arg,sizeof(comedi_trig)))
		return -EFAULT;
	
#if 0
	/* this appears to be the only way to check if we are allowed
	   to write to an area. */
	if(copy_to_user(arg,&it,sizeof(comedi_trig)))
		return -EFAULT;
#endif
	
	if(it.subdev>=dev->n_subdevices)
		return -ENODEV;

	s=dev->subdevices+it.subdev;
	if(s->type==COMEDI_SUBD_UNUSED)
		return -EIO;
	
	/* are we locked? (ioctl lock) */
	if(s->lock && s->lock!=file)
		return -EACCES;

	/* are we busy? */
	if(s->busy)
		return -EBUSY;
	s->busy=file;

	/* make sure channel/gain list isn't too long */
	if(it.n_chan > s->len_chanlist){
		ret = -EINVAL;
		goto cleanup;
	}
	
	/* load channel/gain list */
	/* driver is responsible for freeing chanlist, unless it
	   returns an error */
	user_chanlist=it.chanlist;
	chanlist=kmalloc(s->n_chan*sizeof(int),GFP_KERNEL);
	if(!chanlist){
		ret = -ENOMEM;
		goto cleanup;
	}
	it.chanlist=chanlist;
	
	if(copy_from_user(it.chanlist,user_chanlist,s->n_chan*sizeof(int))){
		ret = -EFAULT;
		goto cleanup;
	}
	
	/* make sure each element in channel/gain list is valid */
	for(i=0;i<it.n_chan;i++){
		if(CR_CHAN(it.chanlist[i])>=s->n_chan){
			ret = -EINVAL;
			goto cleanup;
		}
	}
	if(s->range_type){
		for(i=0;i<it.n_chan;i++){
			if(CR_CHAN(it.chanlist[i])>=s->n_chan ||
		   	CR_RANGE(it.chanlist[i])>=RANGE_LENGTH(s->range_type)){
				ret = -EINVAL;
				goto cleanup;
			}
		}
	}else{
		int chan;

		if(s->range_type_list)
		for(i=0;i<it.n_chan;i++){
			chan=CR_CHAN(it.chanlist[i]);
			if(CR_RANGE(it.chanlist[i])>=RANGE_LENGTH(s->range_type_list[chan])){
				ret = -EINVAL;
				goto cleanup;
			}
		}
	}
	
	s->cur_trig=it;
	
	ret=s->trig(dev,s,&it);
	
	if(ret==0)return 0;  /* driver now owns chanlist */
	if(ret>0)ret=0;
cleanup:

	s->busy=NULL;

	if(chanlist)
		kfree(chanlist);
	
	return ret;
}


/*
	COMEDI_LOCK
	lock subdevice
	
	arg:
		subdevice number
	
	reads:
		none
	
	writes:
		none

	non-RT linux always controls rtcomedi_lock_semaphore.  If an
	RT-linux process wants the lock, it first checks rtcomedi_lock_semaphore.
	If it is 1, it knows it is pre-empting this function, and fails.
	Obviously, if RT-linux fails to get a lock, it *must* allow
	linux to run, since that is the only way to free the lock.
	
	This function is not SMP compatible.

	necessary locking:
	- ioctl/rt lock  (this type)
	- lock while subdevice busy
	- lock while subdevice being programmed
	
*/

volatile int rtcomedi_lock_semaphore=0;

static int do_lock_ioctl(comedi_device *dev,unsigned int arg,void * file)
{
	int ret=0;
	comedi_subdevice *s;
	
	if(arg>=dev->n_subdevices)
		return -EINVAL;
	s=dev->subdevices+arg;
	
	rtcomedi_lock_semaphore=1;
	
	if(s->lock && s->lock!=file){
		ret=-EACCES;
	}else{
		s->lock=file;
	}
	
	rtcomedi_lock_semaphore=0;

	return ret;
}


/*
	COMEDI_UNLOCK
	unlock subdevice
	
	arg:
		subdevice number
	
	reads:
		none
	
	writes:
		none

	This function isn't protected by the semaphore, since
	we already own the lock.
*/
static int do_unlock_ioctl(comedi_device *dev,unsigned int arg,void * file)
{
	comedi_subdevice *s;
	
	if(arg>=dev->n_subdevices)
		return -EINVAL;
	s=dev->subdevices+arg;
	
	if(s->lock && s->lock!=file)
		return -EACCES;
	
	if(s->lock==file)
		s->lock=NULL;

	return 0;
}


#ifdef LINUX_V20

static int comedi_write_v20(struct inode *inode,struct file *file,const char *buf,int nbytes)
{
	return comedi_write_v22(file,buf,nbytes,NULL);
}

static int comedi_read_v20(struct inode *inode,struct file *file,const char *buf,int nbytes)
{
	return comedi_read_v22(file,buf,nbytes,NULL);
}

static int comedi_lseek_v20(struct inode * inode,struct file *file,off_t offset,int origin)
{
	return comedi_lseek_v22(file,offset,origin);
}

static void comedi_close_v20(struct inode *inode,struct file *file)
{
	comedi_close_v22(inode,file);
}

#endif


static ssize_t comedi_write_v22(struct file *file,const char *buf,size_t nbytes,loff_t *offset)
{
#if 0
	int retval;
	kdev_t minor=MINOR(file->f_dentry->d_inode->i_rdev);
	comedi_device *dev=comedi_devices+minor;
#endif

	if(!nbytes)return 0;

	/* read data */
	
	return -EIO;
}

static ssize_t comedi_read_v22(struct file * file,char *buf,size_t nbytes,loff_t *offset)
{
#if 0
	int retval=0;
	kdev_t minor=MINOR(file->f_dentry->d_inode->i_rdev);
	comedi_device *dev=comedi_devices+minor;
	int n,count=0;
	struct wait_queue wait={current,NULL};
#endif

	if(!nbytes)return 0;

	return -EIO;
}

#if 0
/*
	the 0.5 series comedi_read()
*/
static int comedi_read(struct inode *inode,struct file * file,char *buf,int nbytes)
{
	kdev_t minor=MINOR(file->f_dentry->d_inode->i_rdev);
	comedi_device *dev=comedi_devices+minor;
	int n,count=0,retval=0;
	struct wait_queue wait={current,NULL};

	if(!file->private_data)return -EIO;
	if(!nbytes)return 0;

	add_wait_queue(&dev->rdwait,&wait);
	current->state=TASK_INTERRUPTIBLE;
	while(nbytes>0){
		n=copy_from_buffer((comedi_bufptr *)(file->private_data),buf+count,nbytes);
		if(!n){
			if(file->f_flags&O_NONBLOCK){
				retval=-EAGAIN;
				break;
			}
			if(signal_pending(current)){
				retval=-ERESTARTSYS;
				break;
			}
			schedule();
			continue;
		}
		count+=n;
		nbytes-=n;
	}
	current->state=TASK_RUNNING;
	remove_wait_queue(&dev->rdwait,&wait);

	return (count ? count : retval);
}
#endif


static loff_t comedi_lseek_v22(struct file *file,loff_t offset,int origin)
{
	return -ESPIPE;
}

static int comedi_open(struct inode *inode,struct file *file)
{
	kdev_t minor=MINOR(inode->i_rdev);
	comedi_device *dev=comedi_devices+minor;

	if(minor>=COMEDI_NDEVICES)return -ENODEV;
	if(!comedi_devices[minor].rem && !suser())
		return -ENODEV;
	MOD_INC_USE_COUNT;
	comedi_devices[minor].use_count++;
	
	file->private_data=NULL;

	return 0;
}

static int comedi_close_v22(struct inode *inode,struct file *file)
{
	kdev_t minor=MINOR(inode->i_rdev);
	comedi_device *dev=comedi_devices+minor;
	comedi_subdevice *s;
	int i;

	for(i=0;i<dev->n_subdevices;i++){
		s=dev->subdevices+i;
		
		if(s->busy==file){
			/* do something */
		}
		if(s->lock==file){
			s->lock=NULL;
		}
	}
	
	MOD_DEC_USE_COUNT;
	dev->use_count--;
	
	return 0;
}


/*
	kernel compatibility
*/

#ifdef LINUX_V20
static struct file_operations comedi_fops={
	lseek		= comedi_lseek_v20,
	ioctl		= comedi_ioctl_v20,
	open		= comedi_open_v20,
	close		= comedi_close_v20,
};
#endif

#ifdef LINUX_V22
#define comedi_ioctl_v22 comedi_ioctl
#define comedi_open_v22 comedi_open

static struct file_operations comedi_fops={
	llseek		: comedi_lseek_v22,
	ioctl		: comedi_ioctl_v22,
	open		: comedi_open_v22,
	release		: comedi_close_v22,
	read		: comedi_read_v22,
	write		: comedi_write_v22,
};
#endif


extern struct symbol_table comedi_syms;

int init_module(void)
{
	printk("comedi: version " COMEDI_VERSION " - David Schleef <ds@stm.lbl.gov>\n");
	if(register_chrdev(COMEDI_MAJOR,"comedi",&comedi_fops)){
		printk("comedi: unable to get major %d\n",COMEDI_MAJOR);
		return -EIO;
	}
	comedi_devices=(comedi_device *)kmalloc(sizeof(comedi_device)*COMEDI_NDEVICES,GFP_KERNEL);
	if(!comedi_devices)
		return -ENOMEM;
	memset(comedi_devices,0,sizeof(comedi_device)*COMEDI_NDEVICES);
	init_polling();
#ifdef LINUX_V20
#ifndef CONFIG_MODVERSIONS
	register_symtab(&comedi_syms);
#endif
#endif

	/* XXX requires /proc interface */
	comedi_proc_init();
	
	return 0;
}

void cleanup_module(void)
{
	int i;

	if(MOD_IN_USE)
		printk("comedi: module in use -- remove delayed\n");
	comedi_proc_cleanup();
	for(i=0;i<COMEDI_NDEVICES;i++){
		if(comedi_devices[i].rem)
			comedi_devices[i].rem(comedi_devices+i);
	}
	cleanup_polling();
	kfree(comedi_devices);

	unregister_chrdev(COMEDI_MAJOR,"comedi");
}

void report_sample(comedi_device *dev,sampl_t s)
{
/*
	copy_to_buffer(&(dev->rdbuf),s,sizeof(comedi_sample));
	if(dev->chinfo[s->chan].lastdata)
		dev->chinfo[s->chan].lastdata->pval=s->data;
*/
}

void comedi_error(comedi_device *dev,const char *s)
{
	printk("comedi%d: %s: %s\n",dev->minor,dev->driver_name,s);
}

static void postconfig(comedi_device *dev)
{
	int i;
	comedi_subdevice *s;

	for(i=0;i<dev->n_subdevices;i++){
		s=dev->subdevices+i;

		if(s->type==COMEDI_SUBD_UNUSED)
			continue;

		if(s->len_chanlist==0)
			s->len_chanlist=1;
	}

}


#if 0

#ifdef LINUX_V20

/*
   need to generate modversions .def file for symbols if
   we want to use modversions and export them
*/
#ifndef CONFIG_MODVERSIONS

static struct symbol_table comedi_syms = {
#include <linux/symtab_begin.h>
	X(comedi_get),
	X(comedi_put),
#include <linux/symtab_end.h>
};

#endif
#endif

#ifdef LINUX_V22

EXPORT_SYMBOL(comedi_get);
EXPORT_SYMBOL(comedi_put);

#endif

#endif /* 0 */


