/*
    module/rti800.c
    hardware driver for Analog Devices RTI-800/815 board

    COMEDI - Linux Control and Measurement Device Interface
    Copyright (C) 1998 David A. Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <linux/timer.h>
#include <asm/io.h>
#include <comedi_module.h>


#define RTI800_SIZE 16

#define RTI800_CSR 0
#define RTI800_MUXGAIN 1
#define RTI800_CONVERT 2
#define RTI800_ADCLO 3
#define RTI800_ADCHI 4
#define RTI800_DAC0LO 5
#define RTI800_DAC0HI 6
#define RTI800_DAC1LO 7
#define RTI800_DAC1HI 8
#define RTI800_CLRFLAGS 9
#define RTI800_DI 10
#define RTI800_DO 11
#define RTI800_9513A_DATA 12
#define RTI800_9513A_CNTRL 13
#define RTI800_9513A_STATUS 13


/*
 * flags
 */

#define RTI800_BUSY 0x80
#define RTI800_DONE 0x40
#define RTI800_OVERRUN 0x20
#define RTI800_TCR 0x10
#define RTI800_DMA_ENAB 0x08
#define RTI800_INTR_TC 0x04
#define RTI800_INTR_EC 0x02
#define RTI800_INTR_OVRN 0x01

#define Am9513_8BITBUS

#define Am9513_output_control(a)	outb(a,dev->iobase+RTI800_9513A_CNTRL)
#define Am9513_output_data(a)		outb(a,dev->iobase+RTI800_9513A_DATA)
#define Am9513_input_data()		inb(dev->iobase+RTI800_9513A_DATA)
#define Am9513_input_status()		inb(dev->iobase+RTI800_9513A_STATUS)

#include <am9513.h>


comedi_devinit rti800_init;
static int rti800_rem(comedi_device *dev);
#ifdef USE_IRQ
static void rti800_interrupt(int irq,void *dev,struct pt_regs * regs);
#endif

typedef struct {
  enum { adc_diff, adc_pseudodiff, adc_singleended } adc_mux ;
  enum { adc_bipolar10, adc_bipolar5, adc_unipolar10 } adc_range;
  enum { adc_2comp, adc_straight } adc_coding ;
  enum { dac_bipolar10, dac_unipolar10 } dac0_range, dac1_range;
  enum { dac_2comp, dac_straight } dac0_coding, dac1_coding;
} rti800_private;

#define devpriv ((rti800_private *)dev->private)


#define RTI800_TIMEOUT 100

static int rti800_iai(comedi_device *dev,comedi_trig *it)
{
	int i,hi,lo,gain;
	unsigned long flags;

	gain=getparamval(dev->chinfo[it->chan].paramlist,COMEDI_RANGE);
	outb(it->chan|(gain<<5),dev->iobase+RTI800_MUXGAIN);
	save_flags(flags);
	cli();
	outb(0,dev->iobase+RTI800_CONVERT);
	for(i=0;i<RTI800_TIMEOUT;i++){
		if(inb(dev->iobase+RTI800_CSR)&RTI800_DONE)
			break;
	}
	lo=inb(dev->iobase+RTI800_ADCLO);
	hi=inb(dev->iobase+RTI800_ADCHI);
	restore_flags(flags);

	it->data=(hi<<8)|lo;
	it->data&=0xfff;
        if(devpriv->adc_coding == adc_2comp) { it->data ^= 0x800; }

	return 0;
}

static int rti800_ai(comedi_device *dev,comedi_trig *it)
{
	switch(it->trigsrc){
/*
	case COMEDI_TRIGCLK:
		if(dev->busy[0])return -EBUSY;
		devpriv->ntrig=it->n;
		outb(it->chan|DT2814_ENB|(it->trigvar<<5),
			dev->iobase+DT2814_CSR);
		break;
*/
	default:
		return -EINVAL;
	}
	
	return 0;
} 

static int rti800_ao(comedi_device *dev,comedi_trig *it)
{
   int result = 0;
   int chan = it->chan;

   if ((devpriv->adc_mux == adc_diff && chan == 8) || (chan == 16)) {
     int data = it->data;

     if (devpriv->dac0_coding == dac_2comp) { data -= 0x800; }
     outb(data&0xff,dev->iobase+RTI800_DAC0LO);
     outb(data>>8,dev->iobase+RTI800_DAC0HI);
   } else if ((devpriv->adc_mux == adc_diff && chan == 9) || (chan == 17)) {
     int data = it->data;

     if (devpriv->dac1_coding == dac_2comp) { data -= 0x800; }
     outb(data&0xff,dev->iobase+RTI800_DAC1LO);
     outb(data>>8,dev->iobase+RTI800_DAC1HI);
   } else {
     result = -EINVAL;
         }
   return result;
}

static int rti800_di(comedi_device *dev,comedi_trig *it)
{
	it->data=inb(dev->iobase+RTI800_DI);
	
	return 0;
}

static int rti800_do(comedi_device *dev,comedi_trig *it)
{
	outb(it->data^0xff,dev->iobase+RTI800_DO);
	
	return 0;
}

static int rti800_aisp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_RANGE:
		if(it->pval<0 || it->pval>3)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_RANGE,
			0,it->pval);
		return 0;
	default:
		return -EINVAL;
	}
}

/*
	options[1] - 0 for straight binary, 1 for two's comp
*/

int rti800_init(comedi_device *dev,comedi_devconfig *it)
{
#ifdef USE_IRQ
	int irqs,irq;
	long flags;
#endif
	int i;
	int board;
	int rangetype = 0;

	board=-1;
	if(!strcmp("rti800",it->name))board=0;
	if(!strcmp("rti815",it->name))board=1;
	if(board<0)return 0;

	dev->iobase=it->options[0];
	printk("comedi%d: rti800: 0x%04x ",dev->minor,dev->iobase);
	if(check_region(dev->iobase,RTI800_SIZE)<0){
		printk("I/O port conflict\n");
		return -EIO;
	}
	request_region(dev->iobase,RTI800_SIZE,"rti800");
	dev->iosize=RTI800_SIZE;

	if(board==0)dev->board_name="rti800";
	else dev->board_name="rti815";
	

	dev->driver_name="rti800";

        if (it->options[2] == 0) {
          /* Differential */
          dev->chan = 8+2;
        } else {
          /* Pseudo-differential or single-ended */
          dev->chan = 16+2;
        }

        if(board !=0 ) {
          dev->chan += 2;
        }

	dev->private=kmalloc(sizeof(rti800_private)+
		dev->chan*sizeof(comedi_chinfo),GFP_KERNEL);
	if(!dev->private){
		return -ENOMEM;
	}
	memset(dev->private,0,sizeof(rti800_private)+
		dev->chan*sizeof(comedi_chinfo));
	dev->chinfo=dev->private+sizeof(rti800_private);

        devpriv->adc_mux = it->options[2];
        devpriv->adc_range = it->options[3];
        devpriv->adc_coding = it->options[4];
        devpriv->dac0_range = it->options[5];
        devpriv->dac0_coding = it->options[6];
        devpriv->dac1_range = it->options[7];
        devpriv->dac1_coding = it->options[8];

        for(i = 0 ; i < ((devpriv->adc_mux==adc_diff)?8:16) ; i++) {
          printk("rti800: ADC %04x\n", i);
                 dev->chinfo[i].itrig=rti800_iai;
                 dev->chinfo[i].trig=rti800_ai;
                 dev->chinfo[i].sp=rti800_aisp;
          addparam(&dev->chinfo[i].paramlist,
                   COMEDI_BITS, COMEDI_RDONLY, 12);
          addparam(&dev->chinfo[i].paramlist,
                   COMEDI_MAXDATA, COMEDI_RDONLY, 0xfff);
          addparam(&dev->chinfo[i].paramlist,
                   COMEDI_AD, COMEDI_RDONLY, COMEDI_AD_ANALOG);
          addparam(&dev->chinfo[i].paramlist,
                   COMEDI_IO,COMEDI_RDONLY, COMEDI_IO_INPUT);
          addparam(&dev->chinfo[i].paramlist,
                   COMEDI_LASTDATA, COMEDI_RDONLY, 0);
          switch (devpriv->adc_range) {
            case adc_bipolar10:
              rangetype = RANGE_rti800_ai_10_bipolar; break;
            case adc_bipolar5:
              rangetype = RANGE_rti800_ai_5_bipolar; break;
            case adc_unipolar10:
              rangetype = RANGE_rti800_ai_10_unipolar; break;
          }
          addparam(&dev->chinfo[i].paramlist,
                   COMEDI_RANGE_TYPE, COMEDI_RDONLY, rangetype);
          addparam(&dev->chinfo[i].paramlist,
                   COMEDI_RANGE, COMEDI_RDWR, 0);
                 dev->chinfo[i].lastdata=dev->chinfo[i].paramlist;
         }

        if(board==1) {
          printk("rti800: DAC %04x\n", i);
                 dev->chinfo[i].itrig=rti800_ao;
                 dev->chinfo[i].sp=NULL;
                 addparam(&dev->chinfo[i].paramlist,
                   COMEDI_BITS, COMEDI_RDONLY, 12);
          addparam(&dev->chinfo[i].paramlist,
                   COMEDI_MAXDATA, COMEDI_RDONLY, 0xfff);
                 addparam(&dev->chinfo[i].paramlist,
                     COMEDI_AD, COMEDI_RDONLY, COMEDI_AD_ANALOG);
                 addparam(&dev->chinfo[i].paramlist,
                   COMEDI_IO, COMEDI_RDONLY, COMEDI_IO_OUTPUT);
          switch (devpriv->dac0_range) {
            case dac_bipolar10:
              rangetype = RANGE_rti800_ao_10_bipolar; break;
            case dac_unipolar10:
              rangetype = RANGE_rti800_ao_10_unipolar; break;
          }
                 addparam(&dev->chinfo[i].paramlist,
                   COMEDI_RANGE_TYPE, COMEDI_RDONLY, rangetype);
          addparam(&dev->chinfo[i].paramlist,
                   COMEDI_RANGE, COMEDI_RDONLY, 0);
                 i++;

          printk("rti800: DAC %04x\n", i);
          dev->chinfo[i].itrig=rti800_ao;
                 dev->chinfo[i].sp=NULL;
                 addparam(&dev->chinfo[i].paramlist,
                   COMEDI_BITS, COMEDI_RDONLY, 12);
          addparam(&dev->chinfo[i].paramlist,
                   COMEDI_MAXDATA, COMEDI_RDONLY, 0xfff);
                 addparam(&dev->chinfo[i].paramlist,
                   COMEDI_AD, COMEDI_RDONLY, COMEDI_AD_ANALOG);
          addparam(&dev->chinfo[i].paramlist,
                   COMEDI_IO, COMEDI_RDONLY, COMEDI_IO_OUTPUT);
          switch (devpriv->dac1_range) {
            case dac_bipolar10:
              rangetype = RANGE_rti800_ao_10_bipolar; break;
            case dac_unipolar10:
              rangetype = RANGE_rti800_ao_10_unipolar; break;
         }
                 addparam(&dev->chinfo[i].paramlist,
                   COMEDI_RANGE_TYPE, COMEDI_RDONLY, rangetype);
                 addparam(&dev->chinfo[i].paramlist,
                   COMEDI_RANGE, COMEDI_RDONLY, 0);
                 i++;
         }

         dev->chinfo[i].itrig=rti800_di;
         dev->chinfo[i].sp=NULL;
         addparam(&dev->chinfo[i].paramlist,
                COMEDI_BITS, COMEDI_RDONLY, 8);
         addparam(&dev->chinfo[i].paramlist,
                COMEDI_AD, COMEDI_RDONLY, COMEDI_AD_DIGITAL);
         addparam(&dev->chinfo[i].paramlist,
                COMEDI_IO, COMEDI_RDONLY, COMEDI_IO_INPUT);
         i++;

	dev->chinfo[i].itrig=rti800_do;
	dev->chinfo[i].sp=NULL;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,8);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
	i++;

/* don't yet know how to deal with counter/timers */
#if 0
	dev->chinfo[i].trig=rti800_timer;
	dev->chinfo[i].sp=NULL;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,1);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
	i++;
#endif
	
	comedi_initbuf(&dev->rdbuf);

	dev->rem=rti800_rem;
	
	printk("\n");

	return 1;
}


static int rti800_rem(comedi_device *dev)
{
	int i;
	
	printk("comedi%d: rti800: remove\n",dev->minor);
	
	for(i=0;i<dev->chan;i++)
		delparamlist(dev->chinfo[i].paramlist);
	kfree(dev->private);

	if(dev->irq){
		free_irq(dev->irq,dev);
	}
	release_region(dev->iobase,dev->iosize);

	comedi_freebuf(&dev->rdbuf);

	return 0;
}



